/**
 * Event Parser - a utility to parse events from strings
 *
 * An EventParser struct is created from a C string (NULL terminated).  The
 * type of event can then be interogated (the type member), though both 
 * integers and floats are labelled as floats.  The number of elements is
 * the number of fields in an event, entries in a sequence, or key,value pairs
 * in a dictionary.
 *
 * Individual fields can then be retrieved, either as a basic type, or as a
 * reference type - basic types retrieved as reference types have a single 
 * element which must be retrieved using the relevant method.  getRefValue
 * will return NULL if there are no more elements, and the basic type methods
 * will return 0/ false/ empty string.
 * 
 * Dictionary entries must be obtained via the getDictPair function, which
 * sets two pointers to the key and value as reference types.
 * 
 * No type checking is performed on the get methods; the behaviour when there
 * is a type mismatch is undefined.
 *
 * There are two variants of the get methods; the Next set maintain a counter
 * while those with an idx argument retrieve a specific element, which does not
 * affect the next counter. The next counter cannot be reset.
 *
 * Note that the implementation is lazy - it only parses stuff if requested 
 * to. (though getNumberElements will parse all of that entity). Parsing is
 * cached, so calling getNumberElements repeatedly is cheap, and getRefValue/ 
 * getString/ getDictPair will return the same objects if called repeatedly 
 * with the same index.
 *
 * Where methods return a pointer (getString and getRefValue), that pointer is
 * valid until the top-level AP_EventParser is destroyed by AP_EventParser_dtor
 * (no need to destroy any of its sub- objects).  A well-formed program should
 * have a matching number of AP_EventParser_ctor and AP_EventParser_dtor calls,
 * and only use string or refValue pointers before AP_EventParser_dtor is 
 * called.
 *
 * $Copyright(c) 2007, 2008-2009, 2012-2013 Progress Software Corporation (PSC). All rights reserved.$
 * $Copyright (c) 2013, 2015, 2017 Software AG, Darmstadt, Germany and/or Software AG USA Inc., Reston, VA, USA, and/or its subsidiaries and/or its affiliates and/or their licensors.$
 * Use, reproduction, transfer, publication or disclosure is prohibited except as specifically provided for in your License Agreement with Software AG
 *
 * $Revision: 311836 $
 */
#ifndef AP_EVENTPARSER
#define AP_EVENTPARSER

#include <AP_Types.h>
#include <AP_EventWriter.h>
#include <IAF_Platform.h>

#ifdef __cplusplus
extern "C" {
#endif  /* __cplusplus */

	struct AP_EventParser;

	/**
	 * Event writer
	 */
	struct AP_EventParser {
		AP_EventWriter_type type;
		char* name;
		void* values;
		const char *buf;
		int firstFieldIdx;
		int startIdx; // inclusive
		int endIdx; // exclusive
		int next;
		struct AP_EventParser *parent;
		AP_bool invalid;
#ifdef _DEBUG // enable duplicate dictionary key detection only in a debug build
		void* shadow_dict;
#endif // _DEBUG

		// functions; Where they return a pointer, that pointer is valid until the top-level 
		// AP_EventParser is destroyed by AP_EventParser_dtor (no need to destroy any of its
		// sub-objects).
		AP_int64 (*getNextInt)(struct AP_EventParser* ep);
		const char *(*getNextString)(struct AP_EventParser* ep);
		AP_float64 (*getNextFloat)(struct AP_EventParser* ep);
		AP_decimal (*getNextDecimal)(struct AP_EventParser* ep);
		AP_bool (*getNextBoolean)(struct AP_EventParser* ep);
		struct AP_EventParser* (*getNextRefValue)(struct AP_EventParser* ep);
		void (*getNextDictPair)(struct AP_EventParser* ep, struct AP_EventParser** key, struct AP_EventParser** value);
		int (*getNumberElements)(struct AP_EventParser *ep);
		AP_int64 (*getInt)(struct AP_EventParser* ep, int idx);
		const char *(*getString)(struct AP_EventParser* ep, int idx);
		AP_float64 (*getFloat)(struct AP_EventParser* ep, int idx);
		AP_decimal (*getDecimal)(struct AP_EventParser* ep, int idx);
		AP_bool (*getBoolean)(struct AP_EventParser* ep, int idx);
		struct AP_EventParser* (*getRefValue)(struct AP_EventParser* ep, int idx);
		void (*getDictPair)(struct AP_EventParser* ep, int idx, struct AP_EventParser** key, struct AP_EventParser** value);

		/** Returns the string representation of the 'type' of the 'any' value. */
		const char* (*getAnyType)(struct AP_EventParser* ep, int idx);
		const char* (*getNextAnyType)(struct AP_EventParser* ep);
		/** Can be called on an 'optional' value to check if it is empty. */
		AP_bool (*isOptionalEmpty)(struct AP_EventParser* ep);

	};

	typedef struct AP_EventParser AP_EventParser;
	
	/**
	 * Event parser element, for internal use only.
	 */
	typedef struct {
		AP_EventParser* value;
		AP_EventParser* dictValue;
		char *stringValue;
		int startIdx;
		int endIdx;
		int colonIdx;
	} AP_EventParserElement;

	/*
	 * Event writer functions
	 */
	AP_IAFCORE_API AP_EventParser* AP_IAFCORE_CALL AP_EventParser_ctor(const char* buf);
	AP_IAFCORE_API void AP_IAFCORE_CALL AP_EventParser_dtor(AP_EventParser* ep);
	/** Returns the corresponding AP_EventWriter_type for a string representation of the type */
	AP_IAFCORE_API AP_EventWriter_type AP_IAFCORE_CALL AP_EventParser_toType(const char* type);


#ifdef __cplusplus
}  /* extern "C" */
#endif  /* __cplusplus */

#endif
